/***************************************************************************
*   Copyright (C) 2001 by Rick L. Vinyard, Jr.                            *
*   rvinyard@cs.nmsu.edu                                                  *
*                                                                         *
*   This program is free software; you can redistribute it and/or modify  *
*   it under the terms of the GNU Lesser General Public License as        *
*   published by the Free Software Foundation version 2.1.                *
*                                                                         *
*   This program is distributed in the hope that it will be useful,       *
*   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
*   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
*   GNU General Public License for more details.                          *
*                                                                         *
*   You should have received a copy of the GNU Lesser General Public      *
*   License along with this library; if not, write to the                 *
*   Free Software Foundation, Inc.,                                       *
*   51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA              *
***************************************************************************/
#include "endpoint.h"
#include "utility.h"

#include <iostream>

namespace Conexus {

  Endpoint::Endpoint( bool close_on_destruction ) :
      m_read_thread( NULL ),
      m_write_thread( NULL ),
      m_read_delivery_thread( NULL ),
      m_read_terminate( false ),
      m_read_terminate_immediate( false ),
      m_write_terminate( false ),
      m_write_terminate_immediate( false ),
      m_read_timeout( 500000 ),
      m_write_timeout( 0 ),
      m_timestamp( true ),
      m_dispatcher( NULL ),
      m_use_dispatcher( false ),
      m_close_on_destruction( close_on_destruction ),
      m_state( ENDPOINT_CLOSED ),
      m_readable( true ),
      m_writable( true )
 {
  }


  Endpoint::~Endpoint() {
    if ( this->is_running() )
      this->stop();
  }

  bool Endpoint::is_running() {
    return m_read_thread;
  }

  void Endpoint::start( ) {
    this->start( Conexus::use_dispatcher_default() );
  }

  void Endpoint::start( bool use_dispatcher ) {
    if ( m_read_thread )
      return ;

    if ( this->is_closed() )
      this->open();

    m_use_dispatcher = use_dispatcher;
    if ( m_use_dispatcher ) {
      m_dispatcher = new Glib::Dispatcher();
      if ( m_dispatcher && ! m_dispatcher_connection )
        m_dispatcher_connection = m_dispatcher->connect( sigc::mem_fun( *this, &Endpoint::emit_received_data ) );
    }

    m_read_terminate = false;

    m_read_delivery_thread = Glib::Thread::create( sigc::mem_fun( *this, &Endpoint::read_delivery_thread_main ), true );
    m_read_thread = Glib::Thread::create( sigc::mem_fun( *this, &Endpoint::read_thread_main ), true );
  }

  void Endpoint::stop( ) {
    if ( ! m_read_thread )
      return ;

    m_read_terminate = true;
    m_read_delivery_conditional.signal();
    m_read_thread->join();
    m_read_thread = NULL;
    this->read_thread_cleanup();
    m_read_delivery_thread->join();
    m_read_delivery_thread = NULL;
    this->read_delivery_thread_cleanup();
    m_read_terminate = false;
  }

  void Endpoint::set_blocking_write( bool block ) {
    if ( block ) {
      if ( m_write_thread )
        return ;
      m_write_terminate = false;
      m_write_thread = Glib::Thread::create( sigc::mem_fun( *this, &Endpoint::write_thread_main ), true );
    } else {
      if ( ! m_write_thread )
        return ;
      m_write_terminate = true;
      m_write_conditional.signal();
      m_write_thread->join();
      m_write_thread = NULL;
      this->write_thread_cleanup();
    }
  }

  bool Endpoint::is_blocking_write() {
    return m_write_thread;
  }

  void Endpoint::read_thread_main() {
    Data d;

    while ( ! m_read_terminate ) {
      try {
        d = this->read_data( m_read_timeout );
        this->queue_received_data( d );
      }
      catch ( exception::read::timeout ) {
      }
      catch ( exception::read::disconnected ) {
        break;
      }
    }

    // if we broke out due to a read disconnect we need to do
    // an orderly shutdown which means making sure that all reads
    // are delivered and the delivery thread is shutdown and joined
    if ( ! m_read_terminate ) {
      m_read_terminate = true;
      m_read_delivery_conditional.signal();
      this->read_thread_cleanup();
      m_read_delivery_thread->join();
      this->read_delivery_thread_cleanup();
      m_read_terminate = false;
      m_read_thread = NULL;
      m_read_delivery_thread = NULL;
    }
  }

  void Endpoint::read_thread_cleanup() {
    if ( m_dispatcher_connection )
      m_dispatcher_connection.disconnect();
    if ( m_dispatcher ) {
      m_use_dispatcher = false;
      delete m_dispatcher;
      m_dispatcher = NULL;
    }
  }

  void Endpoint::read_delivery_thread_main() {
    Data current;

    // Grab the read queue lock
    m_read_queue_lock.lock();

    // Repeat main loop until terminate is set
    while ( ! m_read_terminate ) {

      // loop in the conditional until there is something to deliver or terminate is set
      while ( !m_read_terminate && m_read_queue.empty() )
        m_read_delivery_conditional.wait( m_read_queue_lock );

      // if we have a terminate and are set for immediate terminates then break out of the loop
      if ( m_read_terminate && m_read_terminate_immediate ) {
        m_read_queue_lock.unlock();
        break;
      }

      // Otherwise it's time to deliver the data
      while ( ! m_read_queue.empty() ) {
        current = m_read_queue.front();
        m_read_queue.pop();
        m_read_queue_lock.unlock();
        m_signal_data.emit( current );
        m_signal_data_received.emit( current.size() );
        m_read_queue_lock.lock();
      }
    }

    m_read_queue_lock.unlock();
  }

  void Endpoint::read_delivery_thread_cleanup() { }

  void Endpoint::write_thread_main() {
    Data current;
    bool sent;
    while ( ! m_write_terminate ) {
      m_write_queue_lock.lock();

      while ( ! m_write_terminate && m_write_queue.empty() )
        m_write_conditional.wait( m_write_queue_lock );

      if ( m_write_terminate && m_write_terminate_immediate ) {
        m_write_queue_lock.unlock();
        break;
      }

      while ( ! m_write_queue.empty() ) {
        current = m_write_queue.front();
        m_write_queue.pop();
        m_write_queue_lock.unlock();
        sent = false;
        while ( !sent ) {
          try {
            this->write_data( m_write_timeout, current );
            sent = true;
          }
          catch ( exception::write::timeout ) {
          }
        }
        m_signal_data_transmitted.emit( current.size() );
        m_write_queue_lock.lock();
      }
      m_write_queue_lock.unlock();
    }
  }

  void Endpoint::write_thread_cleanup() {}

  sigc::signal<void, const Data> Endpoint::signal_data() {
    return m_signal_data;
  }

  void Endpoint::queue_received_data( const Data d ) {
    m_read_queue_lock.lock();
    m_read_queue.push( d );
    if ( m_use_dispatcher ) {
      m_read_queue_lock.unlock();
      m_dispatcher->emit();
      m_signal_data_received.emit( d.size() );
    } else {
      m_read_delivery_conditional.signal();
      m_read_queue_lock.unlock();
    }
  }

  void Endpoint::queue_write_data( const Data d ) {
    m_write_queue_lock.lock();
    m_write_queue.push( d );
    m_write_conditional.signal();
    m_write_queue_lock.unlock();
  }

  void Endpoint::emit_received_data( ) {
    Data current;
    m_read_queue_lock.lock();
    while ( ! m_read_queue.empty() ) {
      current = m_read_queue.front();
      m_read_queue.pop();
      m_read_queue_lock.unlock();
      m_signal_data.emit( current );
      m_signal_data_received.emit( current.size() );
      m_read_queue_lock.lock();
    }
    m_read_queue_lock.unlock();
  }

  sigc::signal< void, size_t > Endpoint::signal_data_received( ) {
    return m_signal_data_received;
  }

  sigc::signal< void, size_t > Endpoint::signal_data_transmitted( ) {
    return m_signal_data_transmitted;
  }

  long int Endpoint::read_timeout() {
    return m_read_timeout;
  }

  void Endpoint::set_read_timeout( long int t ) {
    m_read_timeout = t;
  }

  long int Endpoint::write_timeout() {
    return m_write_timeout;
  }

  void Endpoint::set_write_timeout( long int t ) {
    m_write_timeout = t;
  }

  bool Endpoint::read_terminate_immediate() {
    return m_read_terminate_immediate;
  }

  void Endpoint::set_read_terminate_immediate( bool i ) {
    m_read_terminate_immediate = i;
  }

  bool Endpoint::write_terminate_immediate() {
    return m_write_terminate_immediate;
  }

  void Endpoint::set_write_terminate_immediate( bool i ) {
    m_write_terminate_immediate = i;
  }

  bool Endpoint::timestamp_received_data() {
    return m_timestamp;
  }

  void Endpoint::set_timestamp_received_data( bool t ) {
    m_timestamp = t;
  }

  ssize_t Endpoint::write( const void* data, size_t size ) throw ( write_exception ) {
    return this->write( Data( data, size, UNMANAGED ) );
  }

  ssize_t Endpoint::write( const Data data ) throw ( write_exception ) {
    if ( m_write_thread ) {
      this->queue_write_data( data );
      return data.size();
    } else
      return this->write_data( m_write_timeout, data );
  }

  Data Endpoint::read( size_t s ) throw ( read_exception ) {
    Data d = this->read_data( m_read_timeout, s );
    if ( m_read_thread )
      this->queue_received_data( d );
    return d;
  }

ssize_t Endpoint::input_available() {
  // TODO throw something better
  throw;
}

  void Endpoint::change_state( long states ) throw ( state_exception ) {
    if ( states == ENDPOINT_UNCHANGED )
      return ;

    // check for inconsistent states
    if ( states & ENDPOINT_CLOSED && states & ENDPOINT_OPENED )
      throw exception::state::inconsistent();

    // perform the proper actions
    if ( is_open() && states & ENDPOINT_CLOSED ) {
      this->close();
      if ( this->is_open() )
        throw exception::state::failed();
    } else if ( this->is_closed() && states & ENDPOINT_OPENED ) {
      this->open();
      if ( this->is_closed() )
        throw exception::state::failed();
    }
  }

  void Endpoint::set_state_opened() {
    if ( is_open() )
      return ;
    m_state &= ~ENDPOINT_CLOSED;
    m_state |= ENDPOINT_OPENED;
    m_signal_opened.emit();
  }

  void Endpoint::set_state_closed() {
    if ( is_closed() )
      return ;
    m_state &= ~ENDPOINT_OPENED;
    m_state |= ENDPOINT_CLOSED;
    m_signal_closed.emit();
  }

  bool Endpoint::close_on_destruction() const {
    return m_close_on_destruction;
  }

  void Endpoint::set_close_on_destruction( bool value ) {
    m_close_on_destruction = value;
  }

  void Endpoint::close_and_reopen( long new_state ) {
    long prev_state = state();
    change_state( ENDPOINT_CLOSED );
    if ( new_state == ENDPOINT_UNCHANGED )
      change_state( prev_state );
    else
      change_state( new_state );
  }

  sigc::signal<void> Endpoint::signal_opened() {
    return m_signal_opened;
  }

  sigc::signal<void> Endpoint::signal_closed() {
    return m_signal_closed;
  }

  bool Endpoint::is_open() {
    return m_state & ENDPOINT_OPENED;
  }

  bool Endpoint::is_closed() {
    return m_state & ENDPOINT_CLOSED;
  }

  long Endpoint::state() {
    return m_state;
  }


} // namespace Conexus

Conexus::Endpoint& operator<<( Conexus::Endpoint & io, const Conexus::Data d ) {
  io.write( d );
  return io;
}

Conexus::Endpoint::pointer operator<<( Conexus::Endpoint::pointer io, const Conexus::Data d ) {
  if ( ! io )
    throw std::invalid_argument("Conexus: << operator with pointer argument, but pointer is NULL.");
  io->write( d );
  return io;
}

Conexus::Endpoint& operator>>( Conexus::Endpoint & io, Conexus::Data d ) {
  d = io.read();
  return io;
}

Conexus::Endpoint::pointer operator>>( Conexus::Endpoint::pointer io, Conexus::Data d ) {
  if ( ! io )
    throw std::invalid_argument("Conexus: >> operator with pointer argument, but pointer is NULL.");
  d = io->read();
  return io;
}
