/***************************************************************************
 *   Copyright (C) 2001 by Rick L. Vinyard, Jr.                            *
 *   rvinyard@cs.nmsu.edu                                                  *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU Lesser General Public License as        *
 *   published by the Free Software Foundation version 2.1.                *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU Lesser General Public      *
 *   License along with this library; if not, write to the                 *
 *   Free Software Foundation, Inc.,                                       *
 *   51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA              *
 ***************************************************************************/
#include "ipv4_tcp.h"

#include <iostream>

#include <sys/ioctl.h>
#include <linux/sockios.h>

namespace Conexus
{
  namespace IPv4
  {

    TCP::TCP( uint32_t address, uint16_t port ) :
        IP( SOCK_STREAM, IPPROTO_TCP )
    {
      m_remote_address.set_address(address);
      m_remote_address.set_port(port);
    }

    TCP::TCP( const std::string& address, uint16_t port ) :
        IP( SOCK_STREAM, IPPROTO_TCP )
    {
      m_remote_address.set_address(address);
      m_remote_address.set_port(port);
    }

    TCP::TCP( int accepted_fd, bool ) :
        IP( SOCK_STREAM, IPPROTO_TCP )
    {
      m_fd = accepted_fd;
      this->set_state_opened();
      m_state |= SOCKET_BOUND | SOCKET_CONNECTED | SOCKET_ACCEPTED;
    }

    TCP::pointer TCP::create( uint16_t port )
    {
      return pointer( new TCP( INADDR_ANY, port ) );
    }

    TCP::pointer TCP::create( uint32_t address, uint16_t port )
    {
      return pointer( new TCP( address, port ) );
    }

    TCP::pointer TCP::create( const std::string& address, uint16_t port )
    {
      return pointer( new TCP( address, port ) );
    }

    TCP::pointer TCP::create( int accepted_fd, bool b )
    {
      return pointer( new TCP( accepted_fd, b ) );
    }

    TCP::~TCP() throw()
    {}

    ssize_t TCP::input_available()
    {
      int n;
      ioctl( m_fd, SIOCINQ, &n );
      if ( n >= 0 )
        return n;
      return 0;
    }

    int TCP::pending_output_size()
    {
      int rsize;
      int error;
      error = ioctl( m_fd, SIOCOUTQ, &rsize );
      return rsize;
    }

    void TCP::on_local_address_changed( )
    {
      // if this is an accepted socket, there is nothing to do when the local address is changed
      if ( is_accepted() )
        return ;

      // but, if this is a client side socket, and the socket is already bound, close and reopen to
      // same state with new local port and address
      if ( is_bound() )
        close_and_reopen();
    }

    void TCP::on_remote_address_changed( )
    {
      // if this is an accepted socket, there is nothing to do when the remote address is changed
      if ( is_accepted() )
        return ;

      // but, if this is a client side socket, and the socket is already connected, close and reopen to
      // same state with new local port and address
      if ( is_connected() )
        close_and_reopen();
    }

    ssize_t TCP::write_data( long int to, const Data data ) throw ( write_exception )
    {
      if ( ! is_accepted() && ! is_connected() )
        try {
          change_state( SOCKET_CONNECTED );
        } catch ( exception::state::failed ) {
          throw exception::write::not_connected();
        }

      return Socket::write_data( to, data );
    }

    Data TCP::read_data( long int to, size_t s ) throw ( read_exception )
    {
      if ( ! ( is_accepted() || is_bound() ) ) {
        try {
          change_state( SOCKET_BOUND );
        } catch ( exception::state::failed ) {
          throw exception::read::not_bound();
        }
      }

      int fds;
      ssize_t bytes;
      Data d;
      fd_set readfd;
      struct timeval timeout;

      if ( this->is_closed() )
        this->open();

      FD_ZERO( &readfd );
      FD_SET( m_fd, &readfd );
      timeout.tv_sec = 0;
      timeout.tv_usec = to;

      if ( to == 0 )
        fds = ::select( m_fd + 1, &readfd, NULL, NULL, NULL );
      else
        fds = ::select( m_fd + 1, &readfd, NULL, NULL, &timeout );

      if ( fds > 0 ) {

        if ( s == 0 ) {
          s = this->input_available();
          if ( s == 0 )
            throw exception::read::disconnected();
        }

        d = Data( s );

        bytes = ::read( m_fd, d, s );

        if ( bytes == 0 )
          throw exception::read::disconnected();

        if ( bytes > 0 ) {
          if ( m_timestamp )
            d.set_current_time();

          if ( d.size() != ( unsigned ) bytes )
            d.resize( bytes );
          return d;
        } else
          throw_read_exception( errno );
      } else if ( fds == 0 ) {
        throw exception::read::timeout();
      }

      /* else fds < 0 */
      // TODO come back and put in proper select throws
      throw;

    }

  }
}
