/***************************************************************************
 *   Copyright (C) 2001 by Rick L. Vinyard, Jr.                            *
 *   rvinyard@cs.nmsu.edu                                                  *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU Lesser General Public License as        *
 *   published by the Free Software Foundation version 2.1.                *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU Lesser General Public      *
 *   License along with this library; if not, write to the                 *
 *   Free Software Foundation, Inc.,                                       *
 *   51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA              *
 ***************************************************************************/
#include "serial.h"

#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>

namespace Conexus
  {

  Serial::Serial(const std::string& device, long unsigned mode)
  {
    if ( m_mode & SERIAL_UNCHANGED )
      m_mode = m_mode & ~SERIAL_UNCHANGED;
    this->set_device(device);
  }

  Serial::pointer Serial::create(const std::string& device, long unsigned mode)
  {
    return pointer( new Serial( device, mode ) );
  }

  Serial::~Serial()
  {}

  void Serial::open( const std::string name, long unsigned mode ) throw (open_exception)
  {
    int flags=O_NOCTTY|O_NDELAY;

    if ( is_open() )
      change_state(ENDPOINT_CLOSED);

    if ( ! (mode & SERIAL_UNCHANGED) )
    m_mode = mode;

    if ( m_mode & (SERIAL_READ|SERIAL_WRITE) )
      flags |= O_RDWR;
    else if ( m_mode & SERIAL_READ )
      flags |= O_RDONLY;
    else if ( m_mode & SERIAL_WRITE )
      flags |= O_WRONLY;

    m_fd = ::open(name.c_str(), flags);

    if ( m_fd == -1 )
      throw open_exception(errno);

    this->set_state_opened();
  }

  void Serial::open( ) throw( open_exception )
  {
    if (m_device.size() == 0)
      throw(exception::open::no_device_set());
    this->open( m_device );
  }

  long unsigned Serial::mode( )
  {
    return m_mode;
  }

  std::string Serial::device( )
  {
    return m_device;
  }

  void Serial::set_device( const std::string& d, long unsigned mode ) throw (open_exception)
  {
    long unsigned tempstate;

    // if the device is the same and the state is the same, then leave
    if (m_device == d && (m_mode==mode || (mode&SERIAL_UNCHANGED) ))
      return;

    // set the new device
    m_device = d;
    m_mode = mode;

    tempstate = m_state;

    // close the old file (changes the state so this needs to be after check above)
    this->close();

    // only potentially reopen the file if the device is not an empty string
    if ( m_device.size() > 0 )
      this->change_state(tempstate);

    m_signal_device_changed.emit();
  }

  sigc::signal<void> Serial::signal_device_changed()
  {
    return m_signal_device_changed;
  }

}

